var raygun = require('raygun');
const { inspect } = require('util');

const inspectParams = {
  colors: true,
  depth: 8
};

// If used in Sails
var _debug = typeof sails !== 'undefined' ? sails.log.debug : console.log; // eslint-disable-line
var _error = typeof sails !== 'undefined' ? sails.log.error : console.error; // eslint-disable-line
var dEnv = typeof sails !== 'undefined' ? sails.config.environment : null; // eslint-disable-line

// Constants for default configuration
const CONST_statusCodeBadRequest = 400;
const CONST_statusCodeForbidden = 403;
const CONST_statusCodeNotFound = 404;
const CONST_statusCodeServerError = 500;
const CONST_tagDelimiter = ' > ';
const CONST_supportedStatusCodes = [
  CONST_statusCodeBadRequest,
  CONST_statusCodeForbidden,
  CONST_statusCodeNotFound,
  CONST_statusCodeServerError
];

const CONST_statusBadRequest = 'bad request';
const CONST_statusForbidden = 'forbidden';
const CONST_statusNotFound = 'not found';
const CONST_statusServerError = 'server error';

// Raygun Client - protected :)
let raygunClient = null;

function tagsAreincluded(err) {
  if (
    err !== undefined &&
    err !== null &&
    typeof err === 'object' &&
    err.hasOwnProperty('error') &&
    err.hasOwnProperty('tags') &&
    Array.isArray(err.tags)
  ) {
    return true;
  } else {
    return false;
  }
}

const _selfError = (e, str) => {
  if (raygunClient) {
    raygunClient.send(e, {}, null, null, [`errTracker-${str}`]);
  } else {
    _error(`Couldn't report error occured in lib-errTracker to raygun`, e);
    throw e;
  }
};

module.exports = {
  // Attributes
  get raygunClient() {
    return raygunClient;
  },
  env: null,
  badRequest: CONST_statusCodeBadRequest,
  forbidden: CONST_statusCodeForbidden,
  notFound: CONST_statusCodeNotFound,
  serverError: CONST_statusCodeServerError,

  statusBadRequest: CONST_statusBadRequest,
  statusForbidden: CONST_statusForbidden,
  statusNotFound: CONST_statusNotFound,
  statusServerError: CONST_statusServerError,

  /**
   * Initialize the Error Tracker
   * @param  {String} apiKey   The RayGun API Key
   * @param  {Object} options  Example: { env: "prod" }.
   * @return {Object}          Self
   */
  init(apiKey, { env = dEnv, logToConsole = true } = {}) {
    try {
      this.env = env;
      this.logToConsole = logToConsole;
      raygunClient = new raygun.Client().init({
        apiKey,
        filters: ['apiKeyAlyneAdmin']
      });
      return this;
    } catch (e) {
      _selfError(e, 'init');
    }
  },

  /** Logs the error to the error tracking cloud service and to the error log in a standardised way
   *
   * @param {Object|String} error - the error, can be String, errObj or new Error()
   * @param {Number} errorCode - the error code as a number
   * @param {Object} payload - the payload object attached for further details
   * @param {Object} req - the sails req object
   * @param {Array} tagsToAdd - the tags to be added: specifies the place where the error occurred, also env and the errorCode
   * @param {Object|function} cb - the callback that will be called after appending 'tagsToAdd' to the 'error' Object. If null, the error is logged
   */
  trackError(error, errorCode, payload, req, tagsToAdd, cb) {
    try {
      // if (tagsAreincluded(error)) {
      //   _debug('error.tags@trackErrror :', error.tags);
      // }

      var errorMsg = this.getErrorMsg(error, errorCode);
      var errorCodeCompiled = this.getErrorCode(error, errorCode);

      var tags = this.getTags(errorCodeCompiled, tagsToAdd, error);
      var payLoadCompiled = this.getPayload(error, payload);

      let newErrorObject = {
        error: '',
        errorCode: 500,
        data: payLoadCompiled
      };

      if (error) {
        newErrorObject = {
          error: this.getErrorMsg(error, errorCode),
          errorCode: error.errorCode || this.getErrorCode(error, errorCode),
          data: payLoadCompiled
        };
      }

      let payloadStr = inspect(payLoadCompiled, inspectParams);

      if (tagsAreincluded(error)) {
        newErrorObject.tags = tagsToAdd.concat(error.tags);
      } else {
        newErrorObject.tags = tagsToAdd;
      }

      if (cb === null || cb === undefined) {
        if (tagsAreincluded(error)) {
          tags = tags.concat(error.tags);
          tagsToAdd = tagsToAdd.concat(error.tags);
        }
        var logEntry = '';
        if (this.raygunClient) {
          this.raygunClient.user = function(req) {
            if (req && req.userObjectId) {
              return req.userObjectId;
            }
          };

          if (typeof errorMsg === 'string') {
            errorMsg = new Error(errorMsg);
          }

          try {
            this.raygunClient.send(errorMsg, payLoadCompiled, null, req, tags);
          } catch (e) {
            this.raygunClient.send(errorMsg, payloadStr, null, req, tags);
          }
        }

        const consoleErrCompiled = this.getTags(
          errorCodeCompiled,
          tagsToAdd,
          false,
          true
        );
        let consoleErrStr = inspect(consoleErrCompiled, inspectParams);

        // log to console
        logEntry =
          errorMsg +
          ' (' +
          errorCodeCompiled +
          ') ' +
          '|| ' +
          consoleErrStr +
          `\n ${payloadStr}`;

        if (req) {
          logEntry += ' || U: ' + req.userObjectId + ' | O: ' + req.orgObjectId;
        }

        if (this.logToConsole) {
          _error(logEntry);
        }
        return newErrorObject;
      } else {
        return cb(newErrorObject, null);
      }
    } catch (e) {
      _selfError(e, 'trackError');
    }
  },

  /**
   * A wrapper around trackError, It returns a function that throws and tracks if an error is passed to it.
   *
   * @name getThrower
   * @function
   * @param {String} functionId The location of the function using error tracking. It has to be of format <FileName>.<FunctionName>.
   * @returns {Function} Throws an error if this function is passed one. Its arguments: err, errCode, payload, uniqueString in the function to identify the location
   */
  getThrower: function(functionId) {
    if (!functionId) {
      throw 'You have to pass a string.';
    }
    return (err, code, payload, strID) =>
      this.throwIfErr(err, code, payload, [functionId, strID]);
  },

  /**
   * A wrapper around trackError. It throws and tracks if an error is passed to it
   *
   * @name throwIfErr
   * @function
   * @param {Object|String} error The error string or a throwable error or an error Object
   * @param {Number} errorCode the errorCode being assigned. Use null or 400 or 403 or 404 but never 500.
   * @param {Object} payload additional params that would help debugging
   * @param {Array} tagsToAdd The tags that would help locate and trace the error occurred
   */
  throwIfErr: function(err, code, payload, tags) {
    if (err) {
      throw this.trackError(err, code, payload, null, tags);
    }
  },

  /**
   * Pass an error with proper params and get back an error object that is compliant to the alyne tracking paradigm. The error is tracked only if a valid res Object is passed, else this just returns an errorObject. Also if res is oassed, The request is responded with appropriate response like in trackErrorAndRespond
   *
   * @name trackThrowable
   * @function
   * @param {Object|String} error The error string or a throwable error or an error Object
   * @param {Number} errorCode the errorCode being assigned
   * @param {Object} payload additional params that would help debugging
   * @param {Object} req the sails req Object
   * @param {Object} res The resObject. If a valid res is passed, its used to respond with an error(like trackErrorAndRespond). And the error is tracked.
   * @param {array} tagsToAdd The tags that would help locate and trace the error occurred
   * @returns {Object} The throwable error object modified for alyne tracking
   */
  trackThrowable: function(error, errorCode, payload, req, res, tagsToAdd) {
    try {
      var errorMsg = this.getErrorMsg(error, errorCode);
      var errorCodeCompiled = this.getErrorCode(error, errorCode);

      var tags = this.getTags(errorCodeCompiled, tagsToAdd, error);
      var payLoadCompiled = this.getPayload(error, payload);

      let newErrorObject = {
        error: '',
        errorCode: 500,
        data: payLoadCompiled
      };

      if (error) {
        newErrorObject = {
          error: this.getErrorMsg(error, errorCode),
          errorCode: error.errorCode || this.getErrorCode(error, errorCode),
          data: payLoadCompiled
        };
      }

      let payloadStr = inspect(payLoadCompiled, inspectParams);

      if (tagsAreincluded(error)) {
        newErrorObject.tags = tagsToAdd.concat(error.tags);
      } else {
        newErrorObject.tags = tagsToAdd;
      }
      if (res) {
        if (tagsAreincluded(error)) {
          tags = tags.concat(error.tags);
          tagsToAdd = tagsToAdd.concat(error.tags);
        }
        var logEntry = '';
        if (this.raygunClient) {
          this.raygunClient.user = function(req) {
            if (req && req.userObjectId) {
              return req.userObjectId;
            }
          };

          if (typeof errorMsg === 'string') {
            errorMsg = new Error(errorMsg);
          }

          try {
            this.raygunClient.send(errorMsg, payLoadCompiled, null, req, tags);
          } catch (e) {
            this.raygunClient.send(errorMsg, payloadStr, null, req, tags);
          }
        }

        // log to console
        logEntry =
          errorMsg +
          ' (' +
          errorCodeCompiled +
          ') ' +
          '|| ' +
          inspect(
            this.getTags(errorCodeCompiled, tagsToAdd, false, true),
            inspectParams
          ) +
          `\n ${payloadStr}`;

        if (req) {
          logEntry += ' || U: ' + req.userObjectId + ' | O: ' + req.orgObjectId;
        }

        if (this.logToConsole) {
          _error(logEntry);
        }

        var errRes = {};
        errRes.doNotLog = true;
        errRes.msg = this.getErrorMsg(error, errorCode);

        // Send res by switch
        switch (errorCode) {
          case this.badRequest:
            return res.badRequest(errRes);

          case this.forbidden:
            return res.forbidden(errRes);

          case this.notFound:
            return res.notFound(errRes);

          default:
            return res.serverError(errRes);
        }
      }
      return newErrorObject;
    } catch (e) {
      _selfError(e, 'trackThrowable');
    }
  },

  trackErrorAndRespond(error, errorCode, payload, req, res, tagsToAdd) {
    try {
      errorCode = this.getErrorCode(error, errorCode);

      this.trackError(error, errorCode, payload, req, tagsToAdd);

      // Add do not report flag to res

      var errRes = {};
      errRes.doNotLog = true;
      errRes.msg = this.getErrorMsg(error, errorCode);

      // Send res by switch
      switch (errorCode) {
        case this.badRequest:
          return res.badRequest(errRes);

        case this.forbidden:
          return res.forbidden(errRes);

        case this.notFound:
          return res.notFound(errRes);

        default:
          return res.serverError(errRes);
      }
    } catch (e) {
      _selfError(e, 'trackErrorAndRespond');
    }
  },

  /** Extracts the best error code retrievable from various error types: String, errObj or new Error()
   *
   * @param {Object|String} error - the error, can be String, errObj or new Error()
   * @param {Number} errorCode - the error code as a number
   */

  getErrorCode(error, errorCode) {
    try {
      var newErrorCode = this.serverError;

      if (typeof errorCode === 'number') {
        if (errorCode === 401) {
          newErrorCode = this.forbidden;
        } else if (CONST_supportedStatusCodes.indexOf(errorCode) > -1) {
          newErrorCode = errorCode;
        }
      } else if (typeof errorCode === 'string') {
        newErrorCode = this.convertErrorCode(errorCode);
      } else {
        // Respect errObj
        if (error && error.hasOwnProperty('status')) {
          newErrorCode = this.convertErrorCode(error.status);
        }

        if (error && error.hasOwnProperty('errorCode')) {
          newErrorCode = error.errorCode;
        }
      }

      return newErrorCode;
    } catch (e) {
      _selfError(e, 'getErrorCode');
    }
  },

  /**
   * Converts one error code from String to Number notation or vice versa based on param errorCode
   * Defaults to 500 (Number)
   * @param  {String|Int} errorCode the error code as String or Number to convert
   * @return {String|Int}           the converted error code as String or Number
   */
  convertErrorCode(errorCode) {
    try {
      var newErrorCode;

      switch (errorCode) {
        // Translate String to Number codes
        case CONST_statusServerError:
          newErrorCode = this.serverError;
          break;

        case CONST_statusBadRequest:
          newErrorCode = this.badRequest;
          break;

        case CONST_statusForbidden:
          newErrorCode = this.forbidden;
          break;

        case CONST_statusNotFound:
          newErrorCode = this.notFound;
          break;

        // Translate Number to String codes
        case this.badRequest:
          newErrorCode = CONST_statusCodeBadRequest;
          break;

        case this.forbidden:
          newErrorCode = CONST_statusCodeForbidden;
          break;

        case this.notFound:
          newErrorCode = CONST_statusCodeNotFound;
          break;

        case this.serverError:
          newErrorCode = CONST_statusCodeServerError;
          break;

        default:
          newErrorCode = CONST_statusCodeServerError;
          break;
      }

      return newErrorCode;
    } catch (e) {
      _selfError(e, 'convertErrorCode');
    }
  },

  /** Extracts the best error msg retrievable from various error types: String, errObj or new Error()
   *
   * @param {Object|String} error - the error, can be String, errObj or new Error()
   * @param {Number} errorCode - the error code as a number
   ¥*/

  getErrorMsg(error, errorCode) {
    try {
      if (typeof error === 'string') {
        return error;
      }

      var errorMsg = '';

      if (error === null || typeof error === 'undefined') {
        errorMsg = this.getErrorCode(error, errorCode);
      } else if (typeof error === 'number') {
        errorMsg = error;
      } else if (typeof error === 'object') {
        // Respect errObj
        if (error.hasOwnProperty('status')) {
          if (error.hasOwnProperty('msg')) {
            errorMsg = error.msg;
          } else {
            errorMsg = this.getErrorCode(error, errorCode);
          }
        }
        if (error.hasOwnProperty('error')) {
          errorMsg = error.error;
        } else {
          return error;
        }
      } else {
        // Try to return it directly
        return error;
      }

      return errorMsg;
      // return new Error(errorMsg);
    } catch (e) {
      _selfError(e, 'getErrorMsg');
    }
  },

  getPayload(error, payload) {
    try {
      const newPayload =
        error && error.data
          ? Array.isArray(error.data) ? error.data : [error.data]
          : [];
      if (payload) {
        newPayload.push(payload);
      }
      return newPayload;
    } catch (e) {
      _selfError(e, 'getPayload');
    }
  },

  getTags(errorCode, tags, error, listErrorCode, printFriendly) {
    try {
      //var newTags = _.union(['env-' + sails.config.environment], tags);
      var newTags = (this.env ? ['env-' + this.env] : []).concat(tags);

      listErrorCode = typeof listErrorCode === 'undefined' || listErrorCode;
      printFriendly = typeof printFriendly !== 'undefined' || printFriendly;

      if (listErrorCode) {
        //newTags = _.union( newTags, [errorCode]);
        newTags = newTags.concat([errorCode]);
      }

      if (printFriendly) {
        return newTags.join(CONST_tagDelimiter);
      } else {
        return newTags;
      }
    } catch (e) {
      _selfError(e, 'getTags');
    }
  }
};
